/*
==========================================================================

	auto-config.c	
	
		 Douglas Yarrington		(yarri@ee.eng.ohio-state.edu)
		
		March 1, 1992
		Release Version	0.1d1		
		
		Developed as part of the FilterTop project to provide filter
		authors with a simple method of user configuration manipulation.
		
		This code requires two external functions: mystrtol and mystrchr.
		
		There are two resources used, fCFG and fVLD. fCFG is the general
		configuration data resouce, used by all filters. fVLD is the
		resourced used by auto-config filters to specify the nature of
		the configuration dialog. The format of the two resouces is
		as follows:
		
			fCFG
			====
			acVers		HLNG
			numPrefs	OCNT
			*****		LSTC
			prefType	TNAM	For GetNamedResouce type function
			prefData	LSTR	long count, followed by byte data
			*****		LSTE

			fVLD		Note: This was formerly SANE, but that's a bit
			====				confusing so I renamed it.
			acVers		HLNG
			numSane		OCNT
			*****		LSTC
			ItemNum		DBYT	Item number corresponds to DITL
			ItemClass	DBYT	Specifies whether an item controls a group
			Group		DBYT	Groups for hiliting
			MinLength	DBYT	Text data min length (count)
			MaxLength	DBYT	Text data max length (count)
			ValidSet	WSTR	Valid characters; null means all
			UpperValue	WSTR	Numerical upper value; null means any
			LowerValue	WSTR	Numerical lower value; null means any
			*****		LSTE
	
	Notes:
		Grouping currently is allowed at one level. That is a checkbox
		which controls the availability of a group of items doesn't have
		another checkbox group within those items.
		
		Grouping exists between radio buttons, and checkboxes which can
		control other checkboxes (but not groups) and radio button groups
		and static and editable text.
		
		Numerical ranges must be specified within the range +/- MAX_LONG;
		It won't handle floating point.

	History:
		920209	DNY	Began development
		920213  DNY Set up access functions
		920220	DNY	Added static text grouping and greying
		920223	DNY	Recoded validity checking routines
		920224	DNY	Changed the grey text routine, still flickers though
		920225	DNY	Changed event handling to remove cursor flashing
					and return/enter interpretation, corrected mystrtol
		920226	DNY	Removed usage of dialog's refCon. 
		920228	DNY	Changed groupMask for simple editText.
		920301	DNY	Added version data and FindConfigData function
		920304	DNY	Changed version data to long
		920304	DNY	Changed handle usage in cfgInit, cfgFinish, and
					cfgFindData

	Status:		
		
				
==========================================================================
*/

/*---------- Defines -----------*/
#define	classGroupHead	1

/*--------- Typedefs -----------*/

typedef struct configData
{
	Byte		itemNum;
	OSType		type;
	Byte		itemClass;
	Byte		group;
	long		dataLen;
	char		*data;
	Byte		minLength;
	Byte		maxLength;
	short		vsLen;
	char		*ValidSet;
	short		upLen;
	char		*UpperValue;
	short		loLen;
	char		*LowerValue;
	struct configData	*next;
} configData;


/*-------- Prototypes ----------*/

// cfgInit support...
configData	*GetConfigData (Handle	fCFGh, Handle fVLDh);
void FillConfigData (configData	*cD, char *fCFGh, char *fVLDh, short itemCount);

// cfgHandleEvent support...
void MaintainGroupConfig (DialogPtr theD, long cfgPtr);
void MaintainRadioGroup (short item, short group, DialogPtr theD, long cfgPtr);
void MaintainCheckboxGroup (short dItem, DialogPtr theD );
char MaintainValidity (DialogPtr theD, long cfgPtr);
char CheckValidity (DialogPtr theD, configData *cD, Str255 str, long cfgPtr );

// cfgFinish support...
void SaveConfigData (DialogPtr theD, Handle fCFGh, long cfgPtr );
void KillConfigData (configData *cD);

// Filter config data access...
// FindConfigData

/*---------- Includes ----------*/

#include "myString.h"
#include "auto-config.h"


/*--------------------- cfgInit ----------------------*/
/*
	Called by FT to initialize data structures that
	are internal to this auto-config code. The data
	is ordered as a linked list, and attached to the
	refCon of the dialog ptr. This then allows for
	multiple confing dlog's at a time.
*/
void cfgInit (DialogPtr theD,Handle fCFGh,Handle fVLDh,long *cfgPtr)
{
short			itemType;
Rect		itemRect;
Handle		itemHandle;
configData	*cD;
char		*istr, istrcnt, hState1, hState2;
long		refCon;

	// Attach cD to the RefCon of this dialog window for later use
	hState1 = HGetState ( fCFGh);
	hState2 = HGetState ( fVLDh);
	HLock(fCFGh);
	HLock(fVLDh);
	*cfgPtr = (long)GetConfigData(fCFGh,fVLDh);
	HSetState ( fCFGh, hState1 );
	HSetState ( fVLDh, hState2 );

	SetPort(theD);
			
	cD = (configData *)*cfgPtr;
	while (cD)
	{
		GetDItem(theD,cD->itemNum,&itemType,&itemHandle,&itemRect);
		switch (itemType)
		{
			// refCons of the controls are set to hold the group that
			// each belongs to, and whether it's on or off:
			// 		refCon = [byte1][byte2][On/Off][Group]
			case 5:		// Checkbox or 
			case 6:		// Radio Button
				refCon = ( cD->data[0]=='1' ? 1 : 0);
				refCon = (refCon << 8 ) | cD->group;
				SetCRefCon ( (ControlHandle)itemHandle, refCon );
			break;
			
			// Install editText with config data
			case 16:	// TextEdit Field
				istrcnt = (char) cD->dataLen;
				istr = (char *) NewPtr ( istrcnt+1 );
				istr[0] = istrcnt;
				BlockMove ( cD->data, &istr[1], istrcnt );
				SetIText ( itemHandle, (StringPtr)istr );
				DisposePtr ( istr );
			break;
		}
		cD = cD->next;
	}
	ShowWindow ( theD );
	
	// Check Group Settings
	MaintainGroupConfig (theD, *cfgPtr);
	// Check validity of editText items
	MaintainValidity (theD, *cfgPtr);
	
}
/*--------------------- cfgInit ----------------------*/

/*------------------ GetConfigData -------------------*/
/*
	Called by cfgInit to parse the fCFG and fVLD 
	resources and fill the configData which will
	be attached to the config dlog's refCon.
*/
configData	*GetConfigData (Handle	fCFGh, Handle fVLDh)
{
long			fCFGvers, fVLDvers;
short			fCFGcount, fVLDcount;
configData	*cD;
char		*fCFGp,*fVLDp;

	fCFGp = *fCFGh;
	fVLDp = *fVLDh;
	BlockMove ( fCFGp, &fCFGvers, 4 );	fCFGp += 4;
	BlockMove ( fCFGp, &fCFGcount, 2 );	fCFGp += 2;
	BlockMove ( fVLDp, &fVLDvers, 4 ); fVLDp += 4;
	BlockMove ( fVLDp, &fVLDcount, 2 ); fVLDp += 2;
	cD = (configData *)NewPtr ( sizeof(configData) );
	// Call recursive fill config routine
	FillConfigData(cD,fCFGp,fVLDp, fCFGcount);
	return (cD);
}
/*------------------ GetConfigData -------------------*/

/*----------------- FillConfigData -------------------*/
/*
	Called by GetConfigData as a recursive method
	of filling the config data structure.
	Data structure is a singly linked list, null
	terminated.
*/
void FillConfigData (configData	*cD, char *fCFGp, char *fVLDp, short itemCount)
{
long	strLen;

	do
	{
		BlockMove ( fCFGp, &(cD->type), 4 );	fCFGp += 4; 
		BlockMove ( fCFGp, &(cD->dataLen), 4 );	fCFGp += 4;
			cD->data = (char *) NewPtr( cD->dataLen );
			BlockMove ( fCFGp, cD->data, cD->dataLen ); fCFGp += cD->dataLen;
		BlockMove ( fVLDp, &(cD->itemNum), 1 );		fVLDp += 1;
		BlockMove ( fVLDp, &(cD->itemClass), 1 );	fVLDp += 1;
		BlockMove ( fVLDp, &(cD->group), 1 );		fVLDp += 1;
		BlockMove ( fVLDp, &(cD->minLength), 1 );	fVLDp += 1;
		BlockMove ( fVLDp, &(cD->maxLength), 1 );	fVLDp += 1;
		BlockMove ( fVLDp, &(cD->vsLen), 2 );		fVLDp += 2;
			cD->ValidSet = (char *) NewPtrClear( cD->vsLen + 1 );
			BlockMove ( fVLDp, cD->ValidSet, cD->vsLen ); fVLDp += cD->vsLen;
		BlockMove ( fVLDp, &(cD->upLen), 2 );	fVLDp += 2;
			cD->UpperValue = (char *) NewPtrClear( cD->upLen + 1 );
			BlockMove ( fVLDp, cD->UpperValue, cD->upLen ); fVLDp += cD->upLen;
		BlockMove ( fVLDp, &(cD->loLen), 2 );	fVLDp += 2;
			cD->LowerValue = (char *) NewPtrClear( cD->loLen + 1 );
			BlockMove ( fVLDp, cD->LowerValue, cD->loLen ); fVLDp += cD->loLen;
		cD->next = 0L;	// Null terminated
		if (--itemCount)
		{
			cD->next = (configData *)NewPtr ( sizeof(configData) );
			cD = cD->next;
		}
	} while (itemCount);
	
}
/*----------------- FillConfigData -------------------*/



/*----------------- cfgHandleEvent -------------------*/
/*
	Called by FT to respond to a dialog event. This
	routine calls DialogSelect so FT doesn't have
	to. This routine must be passed a dlog which
	was initialized by cfgInit; it must have the
	refCon set to the configData structure.
	
	For multiple modeless dlog's, this function could
	be changed so that it checks for a valid dlogptr,
	and if it doesn't get it to skip the event.
*/
state cfgHandleEvent (DialogPtr theD, EventRecord *evt,long cfgPtr)
{
char 			theChar;
short			itemHit,iType;
short				dItem, iCtlPart;
ControlHandle	theControl,iHandle;
Rect			iRect;
Point			where;
long			refCon;
DialogPtr		whichD;
short			itemType;
Rect			itemRect;
Handle			itemHandle;

	switch (evt->what) 
	{
		case keyDown:
			theChar = evt->message & charCodeMask;
			if ((theChar==0x03) || (theChar==0x0D)) 
				if ( MaintainValidity(theD,cfgPtr) )
				{
				long	d;
					GetDItem ( theD, 1, &iType, (Handle *)&iHandle, &iRect );
					HiliteControl ( iHandle, 1 );
					Delay(2,&d);
					HiliteControl ( iHandle, 0 );
					return cfgStOk;
				}
				else return 0;
		break;
			
		case mouseDown:
			where = evt->where;
			GlobalToLocal(&where);
			dItem = FindDItem ( theD, where )+1;
			if ( dItem > 3 )
			{
				iCtlPart = FindControl ( where, (WindowPtr)theD, &theControl );
				if (iCtlPart)
				{
					if (TrackControl(theControl, where, 0L))
						{
							GetDItem ( theD, dItem, &iType, (Handle *)&iHandle, &iRect );
							if (iType == 6)	// radio button
							{
								refCon = GetCRefCon ( (ControlHandle) iHandle );
								MaintainRadioGroup ( dItem, (refCon&0xFF), theD, cfgPtr );
							}
							if (iType == 5) // checkbox
								MaintainCheckboxGroup ( dItem, theD );
							MaintainGroupConfig (theD,cfgPtr );
							MaintainValidity (theD,cfgPtr); 	// set Ok & Save buttons if valid text
						}
				}
			}
		break;
			
		case updateEvt:
			DialogSelect ( evt, &whichD, &itemHit );
			MaintainGroupConfig (theD,cfgPtr); // need to check groups
			MaintainValidity (theD,cfgPtr);	// set Ok & Save buttons
		break;
	}
	
	if ( DialogSelect ( evt, &whichD, &itemHit ) ) 
		if (whichD == theD)
		{
			switch (itemHit)
			{
				case 1:
					return cfgStOk;
				break;

				case 2:
					return cfgStCancel;
				break;

				case 3:
					return cfgStSave;
				break;
				
			}
			MaintainValidity (theD,cfgPtr); 	// set Ok & Save buttons if valid text
		}

	return cfgStNone;

}
/*----------------- cfgHandleEvent -------------------*/

/*----------------- MaintainValidity -----------------*/
/*
	Called by cfgHandleEvent to check the validity of
	the editText items, it they're aren't valid, then
	the Ok and Save buttons are dimmed.
*/
char MaintainValidity (DialogPtr theD, long cfgPtr)
{
short			itemType;
Rect		itemRect;
Handle		itemHandle;
configData	*cD;
Str255		istr;
char		error=0;

	cD = (configData *)cfgPtr;
	while (cD && !error)
	{
		GetDItem(theD,cD->itemNum,&itemType,&itemHandle,&itemRect);
		if (itemType==16)	// editText
		{
			GetIText ( itemHandle, istr );
			cD->dataLen = istr[0];
			DisposePtr ( cD->data );
			cD->data = (char *)NewPtr( istr[0] );
			BlockMove ( &istr[1], cD->data, istr[0] );
			error = CheckValidity(theD, cD, istr, cfgPtr );
		}
		cD = cD->next;
	}
	
	if ( error )
	{
		// Set hilite to disable
		GetDItem(theD,3,&itemType,&itemHandle,&itemRect);
		HiliteControl ( (ControlHandle) itemHandle, 255 );
		GetDItem(theD,1,&itemType,&itemHandle,&itemRect);
		HiliteControl ( (ControlHandle) itemHandle, 255 );
		PenSize(3,3);
		PenPat ( gray );	// dissabled frame
		InsetRect(&itemRect,-4,-4);
		FrameRoundRect(&itemRect,16,16);	// frame OK button
		PenPat (black);
	}
	else
	{
		// Set hilite to enable
		GetDItem(theD,3,&itemType,&itemHandle,&itemRect);
		HiliteControl ( (ControlHandle) itemHandle, 0 );
		GetDItem(theD,1,&itemType,&itemHandle,&itemRect);
		HiliteControl ( (ControlHandle) itemHandle, 0 );
		PenSize(3,3);
		InsetRect(&itemRect,-4,-4);
		FrameRoundRect(&itemRect,16,16);	// frame OK button
	}

	return (!error);
}
/*----------------- MaintainValidity ------------------*/

/*------------------- CheckValidity -------------------*/
/*
	Called by MaintainValidity to check a specific
	editText string to the config data validity conditions.
	These conditions include the length of the string,
	a valid set of characters which the string may
	contain, and limits to the value that the string 
	may represent.
*/
char CheckValidity (DialogPtr theD, configData *cD, Str255 str, long cfgPtr )
{
short 		itemType;
Rect		itemRect;
Handle		itemHandle;
configData	*theCD;
static char	groupMask[30];	// Note, this limits the number of groups to 29!!!
char		i,strLen;
long		refCon;
char		error=0;

	theCD = (configData *)cfgPtr;
	
	for (i=0;i<30;i++) groupMask[i]=1;	// need to assume groups on
	
	while (theCD)
	{		
		// Look for check boxes, denotes edittext which may
		// be deselected
		GetDItem(theD,theCD->itemNum,&itemType,&itemHandle,&itemRect);
		if (itemType == 5)	// Checkbox 
		{
			refCon = GetCRefCon ( (ControlHandle) itemHandle );
			refCon = (refCon >> 8);
			if (theCD->itemClass == classGroupHead)	// group controller
				groupMask[theCD->group] = (char)refCon;
		}
		theCD = theCD->next;
	}
	
	if (groupMask[cD->group])	// only change if group is selected
	{
		strLen = str[0];
		PtoCstr(str);
		// Need to identify which editText is offending
		if (cD->minLength && (cD->minLength > strLen))	// Not enough characters
		{
			error = 1;
		}
		if ( !error && cD->maxLength && (cD->minLength > strLen))	// Too many characters
		{
			error = 1;
		}
		if ( !error && cD->vsLen)	// Check for valid characters
		{
		short	i=0;
			while (!error && i<strLen)
			{
				error = !mystrchr(cD->ValidSet,(short)str[i++]);
			}
		}
		if ( !error && cD->upLen)	// Check for maximum limit
		{
		long d,dstr;
		char *end;
		
			d = mystrtol ( cD->UpperValue, cD->upLen );
			dstr = mystrtol ( (char *)str, strLen );
			error = dstr > d;
		}
		if ( !error && cD->loLen)	// Check for minimum limit
		{
		long d,dstr;
		char *end;
		
			d = mystrtol ( cD->LowerValue, cD->loLen );
			dstr = mystrtol ( (char *)str, strLen );
			error = dstr < d;
		}
	}
	return error;
}
/*------------------- CheckValidity -------------------*/

/*--------------- MaintainGroupConfig -----------------*/
/*
	Called by cfgHandleEvent to control the display of
	groups controlled by checkboxes. If a checkbox
	which controlls a group is disabled, then all
	items in that group are also disabled.
*/
void MaintainGroupConfig (DialogPtr theD, long cfgPtr)
{
short		itemType;
Rect	itemRect;
Handle	itemHandle;
long	refCon;
configData	*cD, *cDcopy;
static char	groupMask[30];	// Note, this limits the number of groups to 29!!!
char	i;

	cD = (configData *)cfgPtr;
	for (i=0;i<30;i++) groupMask[i]=1;	// need to assume groups are on
	cDcopy = cD;
	while (cD)
	{		
		// Pass 1:	Look for check boxes
		GetDItem(theD,cD->itemNum,&itemType,&itemHandle,&itemRect);
		if (itemType == 5)	// Checkbox 
		{
			refCon = GetCRefCon ( (ControlHandle) itemHandle );
			refCon = (refCon >> 8);
			if (cD->itemClass == classGroupHead)	// group controller
				groupMask[cD->group] = (char)refCon;
			SetCtlValue ( (ControlHandle) itemHandle, (char)refCon );
		}
		cD = cD->next;
	}
	
	cD = cDcopy;
	while (cD)
	{
		// Pass 2:	Disable those items which belong to groups that
		//			aren't selected, and enable those which are.
		GetDItem(theD,cD->itemNum,&itemType,&itemHandle,&itemRect);
		if ( itemType != 5 )
		{
			if (groupMask[cD->group])	// item enabled
			{
				if ( itemType == 6 )
				{
					refCon = GetCRefCon ( (ControlHandle) itemHandle );
					SetCtlValue ( (ControlHandle)itemHandle, (refCon>>8));
					HiliteControl ( (ControlHandle) itemHandle, 0 );
				}
				if ( itemType == 16 )
					ShowDItem ( theD, cD->itemNum );
				if ((itemType&0x0F)==8)	// disabled staticText
				{
				Str255	textStr;
					// enable the static text, it had been greyed out
					SetDItem(theD,cD->itemNum,8,itemHandle,&itemRect);
					GetIText ( itemHandle, textStr );
					TextBox ( &textStr[1], (long)textStr[0], &itemRect, 0 );
				}
			}
			else	// item disabled
			{
				if ( itemType == 6 )
				{
					SetCtlValue ( (ControlHandle)itemHandle, 0);
					HiliteControl ( (ControlHandle) itemHandle, 255 );
				}
				if ( itemType == 16 )
					HideDItem ( theD, cD->itemNum );
				if ((itemType&0x0F)==8)	// enabled staticText
				{
				BitMap	bm,obm;
				Str255	textStr;
					// disable and grey out static text
					SetDItem(theD,cD->itemNum,128+8,itemHandle,&itemRect);
					GetIText ( itemHandle, textStr );
					TextBox ( &textStr[1], (long)textStr[0], &itemRect, 0 );
					bm.bounds = itemRect;
					OffsetRect ( &bm.bounds, -itemRect.left, -itemRect.top );
					bm.rowBytes = ( ( (bm.bounds.right - bm.bounds.left) + 31) / 32) * 4;
					bm.baseAddr = NewPtr(bm.rowBytes * bm.bounds.bottom);
					obm = thePort->portBits;
					SetPortBits (&bm);
					PenPat (gray);
					PaintRect (&bm.bounds);
					PenPat (black);
					SetPortBits (&obm);
					CopyBits ( &bm, &thePort->portBits, &bm.bounds, &itemRect, srcBic, NULL );
					DisposPtr ( bm.baseAddr );
				}
			}
		}
		cD = cD->next;
	}
}
/*--------------- MaintainGroupConfig -----------------*/

/*---------------- MaintainRadioGroup -----------------*/
/*
	Called by cfgHandleEvent when a radio button is
	selected in a mousedown event. The the item
	and the group of that radio button are passed,
	all other radio buttons in that group are
	deselected.
*/
void MaintainRadioGroup (short item, short group, DialogPtr theD, long cfgPtr)
{
short		itemType;
Rect	itemRect;
Handle	itemHandle;
long	refCon;
char	refConGroup;
configData *cD;

		cD = (configData *)cfgPtr;
		while (cD)
		{		
			GetDItem(theD,cD->itemNum,&itemType,&itemHandle,&itemRect);
			if (itemType == 6)	// Radio Button
			{
				refCon = GetCRefCon ( (ControlHandle) itemHandle );
				refConGroup = (char)refCon;
				if ( group == refConGroup )
				{
					refCon = ( (cD->itemNum == item) ? 1 : 0 );
					refCon = ( refCon << 8 ) | refConGroup;
					SetCRefCon ( (ControlHandle) itemHandle, refCon );
				}
			}
			cD = cD->next;
		}
}
/*---------------- MaintainRadioGroup -----------------*/

/*--------------- MaintainCheckboxGroup ---------------*/
/*
	Called by cfgHandleEvent when the user selects
	a checkbox, it changes the checkbox's refCon.
*/
void MaintainCheckboxGroup (short dItem, DialogPtr theD )
{
short		itemType;
Rect	itemRect;
Handle	itemHandle;
long	refCon;
char	group;

	GetDItem ( theD, dItem, &itemType, &itemHandle, &itemRect );
	refCon = GetCRefCon ( (ControlHandle) itemHandle );
	group = (char)refCon;
	refCon = ( refCon >> 8 ? 0 : 1);
	refCon = ( (refCon << 8) | group );
	SetCRefCon ( (ControlHandle) itemHandle, refCon );
}
/*--------------- MaintainCheckboxGroup ---------------*/



/*--------------------- cfgFinish ---------------------*/
/*
	Called by FT when the user specifies either Ok,
	Cancel or Save. If the user specifies Ok or Save
	then the config data fCFGh is rewritten with the
	new data.
	
	FT must distinguish between how to handle saving
	config data for Ok versus Save.
*/
state cfgFinish (DialogPtr theD,state s,Handle fCFGh,long cfgPtr)
{
	if ( s==cfgStSave || s==cfgStOk )
		SaveConfigData( theD, fCFGh, cfgPtr );
	KillConfigData( (configData *)cfgPtr );
	return s;
}
/*--------------------- cfgFinish ---------------------*/

/*------------------ SaveConfigData -------------------*/
/*
	Called by cfgFinish to set up the fCFGh to reflect
	the changes made by the user. First the config
	data structure is set to reflect the current state
	of the controls by referencing their refCons, then
	the fCFGh handle is modified to reflect this change.
*/
void SaveConfigData (DialogPtr theD, Handle fCFGh, long cfgPtr )
{
long		fCFGvers;
short		fCFGcount;
char		*fCFGp,*p,cfgHState;
configData	*cD;
long		refCon, ptrSize;
OSErr		memErr;
short 		itemType;
Rect		itemRect;
Handle		itemHandle;

	cfgHState = HGetState( fCFGh );
	HLock ( fCFGh );
	fCFGp = *fCFGh;
	BlockMove ( fCFGp, &fCFGvers, 4 ); fCFGp += 4;
	BlockMove ( fCFGp, &fCFGcount, 2 );
	HSetState ( fCFGh, cfgHState );
	cD = (configData *)cfgPtr;
	ptrSize = 6;
	while (cD)	// count space that will be needed, and set radio & checkbox data
	{
		GetDItem(theD,cD->itemNum,&itemType,&itemHandle,&itemRect);
		switch (itemType)
		{
			case 5:		// checkboxes
			case 6:		// radio buttons
				refCon = GetCRefCon ( (ControlHandle) itemHandle );
				refCon = (refCon >> 8);
				cD->dataLen = 1;
				cD->data[0] = ( refCon ? '1' : '0' );	// On or Off
			break;
		}
		ptrSize += (cD->dataLen + 8);
		cD = cD->next;
	}
	p = (char *)NewPtrClear( ptrSize );
	fCFGp = p;
	BlockMove ( &fCFGvers, fCFGp, 4 );
	fCFGp += 4;
	BlockMove ( &fCFGcount, fCFGp, 2 );
	fCFGp += 2;
	cD = (configData *)cfgPtr;
	while (cD)
	{
		BlockMove( &(cD->type), fCFGp, 4);
		fCFGp += 4;
		BlockMove( &(cD->dataLen), fCFGp, 4);
		fCFGp += 4;
		BlockMove( cD->data, fCFGp, cD->dataLen );
		fCFGp += cD->dataLen;
		cD = cD->next;
	}
	cfgHState = HGetState ( fCFGh );
	HUnlock ( fCFGh );
	SetHandleSize ( fCFGh, ptrSize );
	HLock ( fCFGh );	// Just call me paraniod
	BlockMove ( p, *fCFGh, ptrSize );
	HSetState ( fCFGh, cfgHState );
}
/*------------------ SaveConfigData -------------------*/


/*------------------ KillConfigData -------------------*/
/*
	Called by cfgFinish to kill of the config data
	structure.
*/
void KillConfigData (configData *cD)
{
configData *old;

	do
	{
		DisposPtr ( cD->data );
		DisposPtr ( cD->ValidSet );
		DisposPtr ( cD->UpperValue );
		DisposPtr ( cD->LowerValue );
		old = cD;
		cD = cD->next;
		DisposPtr ( (Ptr)old );
	} while (cD);
}
/*------------------ KillConfigData -------------------*/


/*-------------------- cfgFindData --------------------*/
pascal void cfgFindData(Handle fCFGh, OSType type, char *data, long *dataLen )
{
char	*p,*d, hState;
short	count;
long	vers,dLen;
OSType	name;

	hState = HGetState ( fCFGh );
	HLock ( fCFGh );
	p = *fCFGh;
	BlockMove ( p, &vers, 4);	p += 4;
	BlockMove ( p, &count, 2);	p += 2;
	while (count--)
	{
		BlockMove ( p, &name, 4 ); p += 4;
		BlockMove ( p, &dLen, 4 ); p += 4;
		if (name == type)	// data is the requested type
		{
			if (data == 0L)	// just return data length
				*dataLen = dLen;
			else
			{
				if (*dataLen > dLen)
					*dataLen = dLen;
				BlockMove ( p, data, *dataLen );
			}
			break;
		}
		p += dLen;
	}
	HSetState ( fCFGh, hState );
}
/*-------------------- cfgFindData --------------------*/
